package com.sf.vertx.utils;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.commons.lang3.StringUtils;

import com.sf.vertx.api.pojo.AppConfig;
import com.sf.vertx.api.pojo.Node;
import com.sf.vertx.api.pojo.RouteContent;
import com.sf.vertx.api.pojo.SacService;
import com.sf.vertx.arithmetic.roundRobin.SacLoadBalancing;
import com.sf.vertx.arithmetic.roundRobin.WeightedRoundRobin;
import com.sf.vertx.service.AppConfigService;
import com.sf.vertx.service.impl.AppConfigServiceImpl;

import io.vertx.core.http.HttpServerRequest;
import io.vertx.core.net.SocketAddress;
import io.vertx.ext.web.handler.HttpException;
import lombok.extern.slf4j.Slf4j;

/***
 * 反向代理工具类
 * 
 * @author xy
 *
 */
@Slf4j
public class ProxyTool {
	public final static Map<Integer, String> _ERROR = new HashMap<>();
	static {
		_ERROR.put(400, "Bad Request");
		_ERROR.put(401, "Unauthorized");
		_ERROR.put(403, "Forbidden");
		_ERROR.put(404, "Not Found");
		_ERROR.put(413, "Request Entity Too Large");
		_ERROR.put(415, "Unsupported Media Type");
		_ERROR.put(500, "Internal Server Error");
		_ERROR.put(502, "Bad Gateway");
		_ERROR.put(503, "Service Unavailable");
		_ERROR.put(504, "Gateway Timeout");
		_ERROR.put(504, "Gateway Timeout");
		_ERROR.put(10000, "无法找到路由地址");
		_ERROR.put(10001, "加解密算法传递错误");
	};

	public static SocketAddress resolveOriginAddress(HttpServerRequest request) {
		String appCode = request.getHeader(AppConfigServiceImpl.getSacAppHeaderKey());
		String appHeaderServiceName = request.getHeader(AppConfigServiceImpl.getAppHeaderServiceName());
		log.info("uri:{}, header appCode:{},appHeaderServiceName:{}", request.uri(), appCode, appHeaderServiceName);
		AppConfig appConfig = AppConfigServiceImpl.getAppConfig(appCode);
		if (appConfig != null) {
			SacService sacService = AppConfigServiceImpl.getSacService(appCode, appHeaderServiceName);
			if (sacService != null) {
				SacLoadBalancing sacLoadBalancing = null;
				// 获取service模式
				if (StringUtils.equals(sacService.getServiceModel(), "NORMAL")
						|| StringUtils.equals(sacService.getServiceModel(), "ROUTE")) {
					sacLoadBalancing = AppConfigServiceImpl.getSacLoadBalancing(appCode, appHeaderServiceName);
				} else if (sacService.getRouteConfig() != null
						&& StringUtils.equals(sacService.getRouteConfig().getRouteType(), "HEADER_ROUTE")) {
					List<Node> nodeList = new ArrayList<>();
					for (RouteContent routeContent : sacService.getRouteConfig().getRouteContent()) {
						// 判断是否uri匹配
						String headerRouteKey = request.getHeader(routeContent.getHeaderKey());
						if (routeContent.getHeaderValues() != null
								&& routeContent.getHeaderValues().contains(headerRouteKey)) {
							Node node = new Node();
							node.setIp(routeContent.getServerAddress().getHost());
							node.setPort(routeContent.getServerAddress().getPort());
							node.setWeight(0);
							node.setProtocol(sacService.getServerAddress().getProtocol());
							nodeList.add(node);
							break;
						}
					}

					if (nodeList.size() > 0) {
						sacLoadBalancing = ProxyTool.roundRobin(nodeList);
					}
				}

				if (sacLoadBalancing == null) {
					log.error("app config error. appCode:{},serviceName:{},RouteType:{}, not find config.", appCode,
							appHeaderServiceName, sacService.getRouteConfig().getRouteType());
					throw new HttpException(10000, _ERROR.get(10000));
				}
				// TODO 区分https、http
				Node node = sacLoadBalancing.selectNode();
				SocketAddress socketAddress = SocketAddress.inetSocketAddress(node.getPort(), node.getIp());
				log.info("sacLoadBalancing address:{},port:{}", socketAddress.host(), socketAddress.port());
				return socketAddress;
			} else {
				log.error("app config error. appCode:{},serviceName:{}, appCode, serviceName not find config.", appCode,
						appHeaderServiceName);
				throw new HttpException(10000, _ERROR.get(10000));
			}
		}
		log.error("app config error. appCode:{},serviceName:{}, appCode, serviceName not find config.", appCode,
				appHeaderServiceName);
		throw new HttpException(10000, _ERROR.get(10000));
	}

	private static boolean regexMatch(String pattern, String target) {
		return Pattern.matches(pattern, target);
	}

	public static SacLoadBalancing roundRobin(List<Node> nodeList) {
		WeightedRoundRobin weightedRoundRobin = new WeightedRoundRobin();
		for (Node node : nodeList) {
			int weight = node.getWeight() != null ? node.getWeight() : 1;
			node.setWeight(weight);
			node.setCurrentWeight(weight);
			node.setEffectiveWeight(weight);
			WeightedRoundRobin.totalWeight += node.getEffectiveWeight();
		}
		weightedRoundRobin.init(nodeList);
		return weightedRoundRobin;
	}
}
